/*
**********************************************************************
--------------
LED-LAMP FADER
--------------

Author: Christoph Brennig
Copyright (c) 2011

This code is written for the Atmel Attiny45 microcontroller, but it
can be ported to other MCUs easily. I think mostly the register names
and fuses are different. The following fuse settings are necessary:
(Attiny45!)

lfuse=0xE2
hfuse=0xDF
efuse=0xFF

The internal RC oscillator is enabled at 8 MHz to use all 5 portpins.
We don't need clock accuracy for this application.


License:
--------

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

**********************************************************************
*/

#ifndef F_CPU
#define F_CPU 8000000
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdint.h>

#define GAMMA_ENABLE 1					// set this to 1 to enable gamma correction

#define MULTI 120						// ramp precision multiplier (max. 127)
#define CMAX 255						// maximum pwm counter value, lower values make lower pwm resolution but higher pwm frequencies
#define TMAX 150						// compare match value for timer1, minimum approx. 150

#define LED_R PB0
#define LED_G PB1
#define LED_B PB2

#define BUTTON1 PB3
#define BUTTON2 PB4

#define ALL 3
#define RED 2
#define GRN 0
#define BLU 1

volatile uint8_t temp = 0;
volatile uint8_t pwmCounter = 0;
volatile uint8_t port = 0;
volatile uint8_t abortFade = 0;

volatile uint8_t input;

volatile uint8_t R_g = 0;
volatile uint8_t G_g = 0;
volatile uint8_t B_g = 0;

volatile uint8_t R_target = 0;
volatile uint8_t G_target = 0;
volatile uint8_t B_target = 0;

volatile uint8_t R = 0;
volatile uint8_t G = 0;
volatile uint8_t B = 0;

volatile int16_t dR = 0;
volatile int16_t dG = 0;
volatile int16_t dB = 0;

volatile int16_t mR = 0;
volatile int16_t mG = 0;
volatile int16_t mB = 0;

volatile uint16_t aR = 127;
volatile uint16_t aG = 127;
volatile uint16_t aB = 127;


// gamma correction
const uint8_t led_gamma[256] PROGMEM = {0,1,1,1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,3,3,3,3,3,3,3,3,4,4,4,4,4,4,4,5,5,5,5,5,5,5,6,6,6,6,6,7,7,7,7,7,7,8,8,8,8,8,9,9,9,9,10,10,10,10,10,11,11,11,11,12,12,12,13,13,13,13,14,14,14,15,15,15,15,16,16,16,17,17,17,18,18,18,19,19,20,20,20,21,21,22,22,22,23,23,24,24,25,25,26,26,26,27,27,28,28,29,30,30,31,31,32,32,33,33,34,35,35,36,37,37,38,38,39,40,41,41,42,43,43,44,45,46,46,47,48,49,50,51,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,71,72,73,74,75,77,78,79,80,82,83,84,86,87,89,90,92,93,95,96,98,99,101,102,104,106,107,109,111,113,115,116,118,120,122,124,126,128,130,132,134,137,139,141,143,145,148,150,153,155,157,160,162,165,168,170,173,176,179,181,184,187,190,193,196,199,202,206,209,212,216,219,222,226,229,233,237,241,244,248,252,255};


/*
#########################################################################
#																		#
#	fade-function:														#
#	--------------														#
#																		#
#	fade( R , G , B , R_time , G_time , B_time );						#
#		  |   |   |     |        |        |								#
#		  |   |   |     |        |        ------- fading time blue		#
#		  |   |   |     |        ---------------- fading time green		#
#		  |   |   |     ------------------------- fading time red		#
#		  |   |   ------------------------------- target value blue		#
#		  |   ----------------------------------- target value green	#
#		  --------------------------------------- target value red		#
#																		#
#	The function exits when all leds reached the target value or		#
#	the global variable abortFade is nonzero.							#
#																		#
#########################################################################
*/

void fade(uint8_t R_t, uint8_t G_t, uint8_t B_t, int16_t R_time, int16_t G_time, int16_t B_time) {

	R_target = R_t;
	G_target = G_t;
	B_target = B_t;

	dR = R_target - R;
	dG = G_target - G;
	dB = B_target - B;

	if(!R_time) R_time++;
	if(!G_time) G_time++;
	if(!B_time) B_time++;

	mR = (MULTI * dR) / R_time;
	mG = (MULTI * dG) / G_time;
	mB = (MULTI * dB) / B_time;

	TCCR0B = (1<<CS00);								// Prescaler: clk/1
	TCCR1 = (1<<CS12) | (1<<CS11) | (1<<CS10);		// Prescaler: clk/64

	uint8_t busy = 1;

	while(busy) {									// wait for the leds to reach the target value
		if(((R == R_target) & (G == G_target) & (B == B_target)) | abortFade) busy = 0;
	}

	TCCR1 = 0;
	TCNT1 = 0;

}

ISR(PCINT0_vect) {									// detect button changes

	input = (PINB >> 3) & 3;
	abortFade = 1;

}


ISR(TIMER0_COMPA_vect) {
	
	temp = SREG;

	if(GAMMA_ENABLE) {
		R_g = pgm_read_byte(&led_gamma[R]);			// gamma correction
		G_g = pgm_read_byte(&led_gamma[G]);
		B_g = pgm_read_byte(&led_gamma[B]);
	} else {
		R_g = R;
		G_g = G;
		B_g = B;
	}

	port = 0;

	if(pwmCounter < R_g) port |= (1<<LED_R);		// pwm-stuff
	if(pwmCounter < G_g) port |= (1<<LED_G);
	if(pwmCounter < B_g) port |= (1<<LED_B);

	PORTB = port;

	if(pwmCounter != CMAX) {
		pwmCounter++;
	} else {
		pwmCounter = 0;
	}

	SREG = temp;
}


ISR(TIMER1_OVF_vect) {

	//LED1
	aR += (mR >= 0) ? mR : -mR;					// rasterization algorithms for each ramp
	if(aR >= MULTI) {							// see the German Wikipedia Article "Rasterung von Linien"
		aR -= MULTI;
		if(mR > 0) R++;
		else if(mR < 0) R--;
	}

	if(R == R_target) {
		aR = 127;
		mR = 0;
	}

	//LED2
	aG += (mG >= 0) ? mG : -mG;
	if(aG >= MULTI) {
		aG -= MULTI;
		if(mG > 0) G++;
		else if(mG < 0) G--;
	}

	if(G == G_target) {
		aG = 127;
		mG = 0;
	}

	//LED3
	aB += (mB >= 0) ? mB : -mB;
	if(aB >= MULTI) {
		aB -= MULTI;
		if(mB > 0) B++;
		else if(mB < 0) B--;
	}

	if(B == B_target) {
		aB = 127;
		mB = 0;
	}

}

int main(void) {

	DDRB = (1<<LED_R) | (1<<LED_G) | (1<<LED_B);
	PORTB = 0;

	TIMSK = (1<<OCIE0A) | (1<<TOIE1);	// Interrupts: Timer0 Compare Match A, Timer1 Overflow
	GIMSK = (1<<PCIE);
	PCMSK = (1<<PCINT3) | (1<<PCINT4);	// Pin Change Interrupt 3 and 4 enabled

	OCR0A = TMAX;
	TCCR0A = (1<<WGM01);				// CTC-Mode

	sei();

	input = (PINB >> 3) & 3;

	_delay_ms(300);								// delay due to heavy contact bounce of my switch

	switch(input) {								// This part fades the leds at a
		case RED:								// medium speed for the first time
			abortFade = 0;						// after turning the lamp on.
			fade(255,0,0,1500,1500,1500);
			break;
		case GRN:
			abortFade = 0;
			fade(0,255,0,1500,1500,1500);
			break;
		case BLU:
			abortFade = 0;
			fade(0,0,255,1500,1500,1500);
			break;
		case ALL:
			abortFade = 0;
			fade(255,0,0,1500,1500,1500);
			break;
	}

	while(1) {

		switch(input) {
			case RED:
				abortFade = 0;
				fade(255,0,0,250,250,250);		// This fading sequences produce a
				fade(245,0,0,250,250,250);		// pulsing light with a linear
				fade(216,0,0,250,250,250);		// interpolated cosine function
				fade(139,0,0,500,500,500);		// with a maximum of 255 and a
				fade(110,0,0,250,250,250);		// minimum of 100. Different values
				fade(100,0,0,250,250,250);		// must be re-calculated manually.
				fade(110,0,0,250,250,250);		// The equation is:
				fade(139,0,0,250,250,250);		// 0.5*((max-min)*cos(alpha)+(max+min))				fade(216,0,0,500,500,500);		// where alpha is 0, 30, 60,120, 150,
				fade(245,0,0,250,250,250);		// 180, 210, 240, 300 and 330 degrees.
				break;
			case GRN:
				abortFade = 0;
				fade(0,255,0,250,250,250);
				fade(0,245,0,250,250,250);
				fade(0,216,0,250,250,250);
				fade(0,139,0,500,500,500);
				fade(0,110,0,250,250,250);
				fade(0,100,0,250,250,250);
				fade(0,110,0,250,250,250);
				fade(0,139,0,250,250,250);
				fade(0,216,0,500,500,500);
				fade(0,245,0,250,250,250);
				break;
			case BLU:
				abortFade = 0;
				fade(0,0,255,250,250,250);
				fade(0,0,245,250,250,250);
				fade(0,0,216,250,250,250);
				fade(0,0,139,500,500,500);
				fade(0,0,110,250,250,250);
				fade(0,0,100,250,250,250);
				fade(0,0,110,250,250,250);
				fade(0,0,139,250,250,250);
				fade(0,0,216,500,500,500);
				fade(0,0,245,250,250,250);
				break;
			case ALL:
				abortFade = 0;
				fade(255,255,0,5000,5000,5000);
				fade(0,255,0,5000,5000,5000);
				fade(0,255,255,5000,5000,5000);
				fade(0,0,255,5000,5000,5000);
				fade(255,0,255,5000,5000,5000);
				fade(255,0,0,5000,5000,5000);
				break;
		}

	}

	return 0;

}
